@@ -7,21 +7,66 @@ Utility functions that do not belong in a separate module.
 """
 import re
 
+from hpack import NeverIndexedHeaderTuple
+
 from .exceptions import ProtocolError, FlowControlError
 
-UPPER_RE = re.compile("[A-Z]")
+UPPER_RE = re.compile(b"[A-Z]")
 
 # A set of headers that are hop-by-hop or connection-specific and thus
 # forbidden in HTTP/2. This list comes from RFC 7540 § 8.1.2.2.
-CONNECTION_HEADERS = set([
-    'connection',
-    'proxy-connection',
-    'keep-alive',
-    'transfer-encoding',
-    'upgrade',
+CONNECTION_HEADERS = {
+    b'connection',
+    b'proxy-connection',
+    b'keep-alive',
+    b'transfer-encoding',
+    b'upgrade',
+}
+
+
+_ALLOWED_PSEUDO_HEADER_FIELDS = {
+    b':method',
+    b':scheme',
+    b':authority',
+    b':path',
+    b':status',
+}
+
+
+_SECURE_HEADERS = frozenset([
+    # May have basic credentials which are vulnerable to dictionary attacks.
+    b'authorization', u'authorization',
+    b'proxy-authorization', u'proxy-authorization',
 ])
 
 
+def secure_headers(headers):
+    """
+    Certain headers are at risk of being attacked during the header compression
+    phase, and so need to be kept out of header compression contexts. This
+    function automatically transforms certain specific headers into HPACK
+    never-indexed fields to ensure they don't get added to header compression
+    contexts.
+
+    This function currently implements two rules:
+
+    - 'authorization' and 'proxy-authorization' fields are automatically made
+      never-indexed.
+    - Any 'cookie' header field shorter than 20 bytes long is made
+      never-indexed.
+
+    These fields are the most at-risk. These rules are inspired by Firefox
+    and nghttp2.
+    """
+    for header in headers:
+        if header[0] in _SECURE_HEADERS:
+            yield NeverIndexedHeaderTuple(*header)
+        elif header[0] in (b'cookie', u'cookie') and len(header[1]) < 20:
+            yield NeverIndexedHeaderTuple(*header)
+        else:
+            yield header
+
+
 def is_informational_response(headers):
     """
     Searches a header block for a :status header to confirm that a given
@@ -79,6 +124,29 @@ def guard_increment_window(current, increment):
     return new_size
 
 
+def authority_from_headers(headers):
+    """
+    Given a header set, searches for the authority header and returns the
+    value.
+
+    Note that this doesn't terminate early, so should only be called if the
+    headers are for a client request. Otherwise, will loop over the entire
+    header set, which is potentially unwise.
+
+    :param headers: The HTTP header set.
+    :returns: The value of the authority header, or ``None``.
+    :rtype: ``bytes`` or ``None``.
+    """
+    for n, v in headers:
+        # This gets run against headers that come both from HPACK and from the
+        # user, so we may have unicode floating around in here. We only want
+        # bytes.
+        if n in (b':authority', u':authority'):
+            return v.encode('utf-8') if not isinstance(v, bytes) else v
+
+    return None
+
+
 def validate_headers(headers):
     """
     Validates a header sequence against a set of constraints from RFC 7540.
@@ -117,8 +185,8 @@ def _reject_te(headers):
     its value is anything other than "trailers".
     """
     for header in headers:
-        if header[0] == 'te':
-            if header[1].lower().strip() != 'trailers':
+        if header[0] == b'te':
+            if header[1].lower().strip() != b'trailers':
                 raise ProtocolError(
                     "Invalid value for Transfer-Encoding header: %s" %
                     header[1]
@@ -151,7 +219,7 @@ def _reject_pseudo_header_fields(headers):
     seen_regular_header = False
 
     for header in headers:
-        if header[0].startswith(':'):
+        if header[0].startswith(b':'):
             if header[0] in seen_pseudo_header_fields:
                 raise ProtocolError(
                     "Received duplicate pseudo-header field %s" % header[0]
@@ -164,6 +232,12 @@ def _reject_pseudo_header_fields(headers):
                     "Received pseudo-header field out of sequence: %s" %
                     header[0]
                 )
+
+            if header[0] not in _ALLOWED_PSEUDO_HEADER_FIELDS:
+                raise ProtocolError(
+                    "Received custom pseudo-header field %s" % header[0]
+                )
+
         else:
             seen_regular_header = True
 
